# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: udfunction.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10udfunction.proto\x12\x0b\x66unction.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto";\n\tEventTime\x12.\n\nevent_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp":\n\tWatermark\x12-\n\twatermark\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"z\n\x05\x44\x61tum\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12*\n\nevent_time\x18\x03 \x01(\x0b\x32\x16.function.v1.EventTime\x12)\n\twatermark\x18\x04 \x01(\x0b\x32\x16.function.v1.Watermark"1\n\tDatumList\x12$\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x12.function.v1.Datum"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\xc3\x01\n\x13UserDefinedFunction\x12\x33\n\x05MapFn\x12\x12.function.v1.Datum\x1a\x16.function.v1.DatumList\x12\x38\n\x08ReduceFn\x12\x12.function.v1.Datum\x1a\x16.function.v1.DatumList(\x01\x12=\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x1a.function.v1.ReadyResponseB<Z:github.com/numaproj/numaflow-go/pkg/apis/proto/function/v1b\x06proto3'
)


_EVENTTIME = DESCRIPTOR.message_types_by_name["EventTime"]
_WATERMARK = DESCRIPTOR.message_types_by_name["Watermark"]
_DATUM = DESCRIPTOR.message_types_by_name["Datum"]
_DATUMLIST = DESCRIPTOR.message_types_by_name["DatumList"]
_READYRESPONSE = DESCRIPTOR.message_types_by_name["ReadyResponse"]
EventTime = _reflection.GeneratedProtocolMessageType(
    "EventTime",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTTIME,
        "__module__": "udfunction_pb2"
        # @@protoc_insertion_point(class_scope:function.v1.EventTime)
    },
)
_sym_db.RegisterMessage(EventTime)

Watermark = _reflection.GeneratedProtocolMessageType(
    "Watermark",
    (_message.Message,),
    {
        "DESCRIPTOR": _WATERMARK,
        "__module__": "udfunction_pb2"
        # @@protoc_insertion_point(class_scope:function.v1.Watermark)
    },
)
_sym_db.RegisterMessage(Watermark)

Datum = _reflection.GeneratedProtocolMessageType(
    "Datum",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATUM,
        "__module__": "udfunction_pb2"
        # @@protoc_insertion_point(class_scope:function.v1.Datum)
    },
)
_sym_db.RegisterMessage(Datum)

DatumList = _reflection.GeneratedProtocolMessageType(
    "DatumList",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATUMLIST,
        "__module__": "udfunction_pb2"
        # @@protoc_insertion_point(class_scope:function.v1.DatumList)
    },
)
_sym_db.RegisterMessage(DatumList)

ReadyResponse = _reflection.GeneratedProtocolMessageType(
    "ReadyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _READYRESPONSE,
        "__module__": "udfunction_pb2"
        # @@protoc_insertion_point(class_scope:function.v1.ReadyResponse)
    },
)
_sym_db.RegisterMessage(ReadyResponse)

_USERDEFINEDFUNCTION = DESCRIPTOR.services_by_name["UserDefinedFunction"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z:github.com/numaproj/numaflow-go/pkg/apis/proto/function/v1"
    _EVENTTIME._serialized_start = 95
    _EVENTTIME._serialized_end = 154
    _WATERMARK._serialized_start = 156
    _WATERMARK._serialized_end = 214
    _DATUM._serialized_start = 216
    _DATUM._serialized_end = 338
    _DATUMLIST._serialized_start = 340
    _DATUMLIST._serialized_end = 389
    _READYRESPONSE._serialized_start = 391
    _READYRESPONSE._serialized_end = 421
    _USERDEFINEDFUNCTION._serialized_start = 424
    _USERDEFINEDFUNCTION._serialized_end = 619
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from typing import TypeVar, Type, List

DROP = b"U+005C__DROP__"
ALL = b"U+005C__ALL__"


M = TypeVar("M", bound="Message")
Ms = TypeVar("Ms", bound="Messages")


class Message:
    def __init__(self, key: str, value: bytes):
        self._key = key or ""
        self._value = value or b""

    def __str__(self):
        return str({self._key: self._value})

    def __repr__(self):
        return str(self)

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> bytes:
        return self._value

    @classmethod
    def to_vtx(cls: Type[M], key: str, value: bytes) -> M:
        return cls(key, value)

    @classmethod
    def to_all(cls: Type[M], value: bytes) -> M:
        return cls(ALL, value)

    @classmethod
    def to_drop(cls: Type[M]) -> M:
        return cls(DROP, b"")


class Messages:
    def __init__(self):
        self._messages = []

    def __str__(self):
        return str(self._messages)

    def __repr__(self):
        return str(self)

    def append(self, message: Message) -> None:
        self._messages.append(message)

    def items(self) -> List[Message]:
        return self._messages

    @classmethod
    def as_forward_all(cls: Type[Ms], value: bytes) -> Ms:
        msgs = cls()
        if value:
            msgs.append(Message.to_all(value=value))
        else:
            msgs.append(Message.to_drop())
        return msgs

    def dumps(self) -> str:
        return str(self)

    def loads(self) -> Ms:
        pass


class Datum:
    """
    Class to define the important information for the event.
    Args:
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
    >>> # Example usage
    >>> from pynumaflow.function import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> d = Datum(value=payload, event_time=t1, watermark=t2)
    """

    def __init__(self, value: bytes, event_time: datetime, watermark: datetime):
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark

    def __str__(self):
        value_string = self._value.decode("utf-8")
        return f"value: {value_string}, event_time: {str(self._event_time)}, watermark: {str(self._watermark)}"

    def __repr__(self):
        return str(self)

    @property
    def value(self):
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self):
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self):
        """Returns the watermark of the event."""
        return self._watermark


class IntervalWindow:
    """Defines the start and end of the interval window for the event."""

    def __init__(self, start: datetime, end: datetime):
        self._start = start
        self._end = end

    def __str__(self):
        return f"start: {self._start}, end: {self._end}"

    def __repr__(self):
        return str(self)

    @property
    def start(self):
        """Returns the start point of the interval window."""
        return self._start

    @property
    def end(self):
        """Returns the end point of the interval window."""
        return self._end


class Metadata:
    """Defines the metadata for the event."""

    def __init__(self, interval_window: IntervalWindow):
        self._interval_window = interval_window

    def __str__(self):
        return f"interval_window: {self._interval_window}"

    def __repr__(self):
        return str(self)

    @property
    def interval_window(self):
        """Returns the interval window for the event."""
        return self._interval_window



import unittest
from datetime import datetime, timezone
from google.protobuf import timestamp_pb2 as _timestamp_pb2


from pynumaflow.sink._dtypes import (
    Datum,
)


def mock_message():
    msg = bytes("test_mock_message", encoding="utf-8")
    return msg


def mock_event_time():
    t = datetime.fromtimestamp(1662998400, timezone.utc)
    return t


def mock_watermark():
    t = datetime.fromtimestamp(1662998460, timezone.utc)
    return t


class TestDatum(unittest.TestCase):
    def test_err_event_time(self):
        ts = _timestamp_pb2.Timestamp()
        ts.GetCurrentTime()
        with self.assertRaises(Exception) as context:
            Datum(sink_msg_id="test_id_0", value=mock_message(), event_time=ts, watermark=ts)
        self.assertEqual(
            "Wrong data type: <class 'google.protobuf.timestamp_pb2.Timestamp'> for Datum.event_time",
            str(context.exception),
        )

    def test_err_watermark(self):
        ts = _timestamp_pb2.Timestamp()
        ts.GetCurrentTime()
        with self.assertRaises(Exception) as context:
            Datum(
                sink_msg_id="test_id_0",
                value=mock_message(),
                event_time=mock_event_time(),
                watermark=ts,
            )
        self.assertEqual(
            "Wrong data type: <class 'google.protobuf.timestamp_pb2.Timestamp'> for Datum.watermark",
            str(context.exception),
        )

    def test_value(self):
        d = Datum(
            sink_msg_id="test_id_0",
            value=mock_message(),
            event_time=mock_event_time(),
            watermark=mock_watermark(),
        )
        self.assertEqual(mock_message(), d.value)
        self.assertEqual(
            "id: test_id_0, value: test_mock_message, "
            "event_time: 2022-09-12 16:00:00+00:00, watermark: 2022-09-12 16:01:00+00:00",
            str(d),
        )
        self.assertEqual(
            "id: test_id_0, value: test_mock_message, event_time: 2022-09-12 16:00:00+00:00, watermark: 2022-09-12 16:01:00+00:00",
            repr(d),
        )

    def test_id(self):
        d = Datum(
            sink_msg_id="test_id_0",
            value=mock_message(),
            event_time=mock_event_time(),
            watermark=mock_watermark(),
        )
        self.assertEqual("test_id_0", d.id)

    def test_event_time(self):
        d = Datum(
            sink_msg_id="test_id_0",
            value=mock_message(),
            event_time=mock_event_time(),
            watermark=mock_watermark(),
        )
        self.assertEqual(mock_event_time(), d.event_time)

    def test_watermark(self):
        d = Datum(
            sink_msg_id="test_id_0",
            value=mock_message(),
            event_time=mock_event_time(),
            watermark=mock_watermark(),
        )
        self.assertEqual(mock_watermark(), d.watermark)


if __name__ == "__main__":
    unittest.main()



import asyncio
import logging
from os import environ

from pynumaflow.function.generated import udfunction_pb2
from pynumaflow.function.generated import udfunction_pb2_grpc

from google.protobuf import empty_pb2 as _empty_pb2

import grpc
from typing import Callable, Any, Iterator

from pynumaflow._constants import (
    FUNCTION_SOCK_PATH,
    DATUM_KEY,
)
from pynumaflow.function import Messages, Datum
from pynumaflow.types import NumaflowServicerContext

if environ.get("PYTHONDEBUG"):
    logging.basicConfig(level=logging.DEBUG)

_LOGGER = logging.getLogger(__name__)

UDFMapCallable = Callable[[str, Datum, Any], Messages]


class UserDefinedFunctionServicer(udfunction_pb2_grpc.UserDefinedFunctionServicer):
    """
    Provides an interface to write a User Defined Function (UDFunction)
    which will be exposed over gRPC.

    Args:
        map_handler: Function callable following the type signature of UDFMapCallable
        sock_path: Path to the UNIX Domain Socket

    Example invocation:
    >>> from pynumaflow.function import Messages, Message, Datum, UserDefinedFunctionServicer
    >>> def map_handler(key: str, datum: Datum) -> Messages:
    ...   val = datum.value
    ...   _ = datum.event_time
    ...   _ = datum.watermark
    ...   messages = Messages()
    ...   messages.append(Message.to_vtx(key, val))
    ...   return messages
    >>> grpc_server = UserDefinedFunctionServicer(map_handler)
    >>> grpc_server.start()
    """

    def __init__(self, map_handler: UDFMapCallable, sock_path=FUNCTION_SOCK_PATH):
        self.__map_handler: UDFMapCallable = map_handler
        self.sock_path = sock_path
        self._cleanup_coroutines = []

    def MapFn(
        self, request: udfunction_pb2.Datum, context: NumaflowServicerContext
    ) -> udfunction_pb2.DatumList:
        """
        Applies a function to each datum element.
        The pascal case function name comes from the generated udfunction_pb2_grpc.py file.
        """
        key = ""
        for metadata_key, metadata_value in context.invocation_metadata():
            if metadata_key == DATUM_KEY:
                key = metadata_value

        msgs = self.__map_handler(
            key,
            Datum(
                value=request.value,
                event_time=request.event_time.event_time.ToDatetime(),
                watermark=request.watermark.watermark.ToDatetime(),
            ),
        )

        datums = []
        for msg in msgs.items():
            datums.append(udfunction_pb2.Datum(key=msg.key, value=msg.value))

        return udfunction_pb2.DatumList(elements=datums)

    def ReduceFn(
        self, request_iterator: Iterator[Datum], context: NumaflowServicerContext
    ) -> udfunction_pb2.DatumList:
        """
        Applies a reduce function to a datum stream.
        The pascal case function name comes from the generated udfunction_pb2_grpc.py file.
        """
        # TODO: implement Reduce function
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> udfunction_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the generated udfunction_pb2_grpc.py file.
        """
        return udfunction_pb2.ReadyResponse(ready=True)

    async def __serve(self) -> None:
        server = grpc.aio.server()
        udfunction_pb2_grpc.add_UserDefinedFunctionServicer_to_server(
            UserDefinedFunctionServicer(self.__map_handler), server
        )
        server.add_insecure_port(self.sock_path)
        _LOGGER.info("Server listening on: %s", self.sock_path)
        await server.start()

        async def server_graceful_shutdown():
            logging.info("Starting graceful shutdown...")
            # Shuts down the server with 5 seconds of grace period. During the
            # grace period, the server won't accept new connections and allow
            # existing RPCs to continue within the grace period.
            await server.stop(5)

        self._cleanup_coroutines.append(server_graceful_shutdown())
        await server.wait_for_termination()

    def start(self) -> None:
        """Starts the server on the given UNIX socket."""
        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(self.__serve())
        finally:
            loop.run_until_complete(*self._cleanup_coroutines)
            loop.close()



import tempfile
import unittest
from datetime import datetime, timezone
from typing import List

from grpc import StatusCode
from grpc_testing import server_from_dictionary, strict_real_time
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as _empty_pb2


from pynumaflow._constants import DATUM_KEY
from pynumaflow.sink import Responses, Datum, Response, UserDefinedSinkServicer
from pynumaflow.sink.generated import udsink_pb2


def udsink_handler(datums: List[Datum]) -> Responses:
    responses = Responses()
    for msg in datums:
        if "err" in msg.value.decode("utf-8"):
            responses.append(Response.as_failure(msg.id, "mock sink message error"))
        else:
            responses.append(Response.as_success(msg.id))
    return responses


def mock_message():
    msg = bytes("test_mock_message", encoding="utf-8")
    return msg


def mock_err_message():
    msg = bytes("test_mock_err_message", encoding="utf-8")
    return msg


def mock_event_time():
    t = datetime.fromtimestamp(1662998400, timezone.utc)
    return t


def mock_watermark():
    t = datetime.fromtimestamp(1662998460, timezone.utc)
    return t


class TestServer(unittest.TestCase):
    def __init__(self, method_name) -> None:
        super().__init__(method_name)

        my_servicer = UserDefinedSinkServicer(udsink_handler)
        services = {udsink_pb2.DESCRIPTOR.services_by_name["UserDefinedSink"]: my_servicer}
        self.test_server = server_from_dictionary(services, strict_real_time())

    def test_is_ready(self):
        method = self.test_server.invoke_unary_unary(
            method_descriptor=(
                udsink_pb2.DESCRIPTOR.services_by_name["UserDefinedSink"].methods_by_name["IsReady"]
            ),
            invocation_metadata={},
            request=_empty_pb2.Empty(),
            timeout=1,
        )

        response, metadata, code, details = method.termination()
        expected = udsink_pb2.ReadyResponse(ready=True)
        self.assertEqual(expected, response)
        self.assertEqual(code, StatusCode.OK)

    def test_forward_message(self):
        event_time_timestamp = _timestamp_pb2.Timestamp()
        event_time_timestamp.FromDatetime(dt=mock_event_time())
        watermark_timestamp = _timestamp_pb2.Timestamp()
        watermark_timestamp.FromDatetime(dt=mock_watermark())

        test_datums = [
            udsink_pb2.Datum(
                id="test_id_0",
                value=mock_message(),
                event_time=udsink_pb2.EventTime(event_time=event_time_timestamp),
                watermark=udsink_pb2.Watermark(watermark=watermark_timestamp),
            ),
            udsink_pb2.Datum(
                id="test_id_1",
                value=mock_err_message(),
                event_time=udsink_pb2.EventTime(event_time=event_time_timestamp),
                watermark=udsink_pb2.Watermark(watermark=watermark_timestamp),
            ),
        ]

        request = udsink_pb2.DatumList(elements=test_datums)

        method = self.test_server.invoke_unary_unary(
            method_descriptor=(
                udsink_pb2.DESCRIPTOR.services_by_name["UserDefinedSink"].methods_by_name["SinkFn"]
            ),
            invocation_metadata={},
            request=request,
            timeout=1,
        )

        response, metadata, code, details = method.termination()
        self.assertEqual(2, len(response.responses))
        self.assertEqual("test_id_0", response.responses[0].id)
        self.assertEqual("test_id_1", response.responses[1].id)
        self.assertTrue(response.responses[0].success)
        self.assertFalse(response.responses[1].success)
        self.assertEqual("", response.responses[0].err_msg)
        self.assertEqual("mock sink message error", response.responses[1].err_msg)
        self.assertEqual(code, StatusCode.OK)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass
from datetime import datetime
from typing import List, TypeVar, Type, Optional

R = TypeVar("R", bound="Response")
Rs = TypeVar("Rs", bound="Responses")


@dataclass
class Response:
    id: str
    success: bool
    err: Optional[str]

    @classmethod
    def as_success(cls: Type[R], id_: str) -> R:
        return Response(id=id_, success=True, err=None)

    @classmethod
    def as_failure(cls: Type[R], id_: str, err_msg: str) -> R:
        return Response(id=id_, success=False, err=err_msg)


class Responses:
    def __init__(self, *responses: R):
        self._responses = list(responses) or []

    def __str__(self):
        return str(self._responses)

    def __repr__(self):
        return str(self)

    def append(self, response: R) -> None:
        return self._responses.append(response)

    def items(self) -> List[R]:
        return self._responses

    def dumps(self) -> str:
        return str(self)


class Datum:
    """
    Class to define the important information for the event.
    Args:
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
    >>> # Example usage
    >>> from pynumaflow.function import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> msg_id = "test_id"
    >>> d = Datum(sink_msg_id=msg_id, value=payload, event_time=t1, watermark=t2)
    """

    def __init__(self, sink_msg_id: str, value: bytes, event_time: datetime, watermark: datetime):
        self._id = sink_msg_id or ""
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark

    def __str__(self):
        value_string = self._value.decode("utf-8")
        return f"id: {self._id}, value: {value_string}, event_time: {str(self._event_time)}, watermark: {str(self._watermark)}"

    def __repr__(self):
        return str(self)

    @property
    def id(self) -> str:
        """Returns the id of the event."""
        return self._id

    @property
    def value(self) -> bytes:
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event."""
        return self._watermark



from pynumaflow.function._dtypes import Message, Messages, Datum
from pynumaflow.function.server import UserDefinedFunctionServicer

__all__ = ["Message", "Messages", "Datum", "UserDefinedFunctionServicer"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: udsink.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0cudsink.proto\x12\x07sink.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto";\n\tEventTime\x12.\n\nevent_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp":\n\tWatermark\x12-\n\twatermark\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"~\n\x05\x44\x61tum\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12&\n\nevent_time\x18\x03 \x01(\x0b\x32\x12.sink.v1.EventTime\x12%\n\twatermark\x18\x04 \x01(\x0b\x32\x12.sink.v1.Watermark\x12\n\n\x02id\x18\x05 \x01(\t"-\n\tDatumList\x12 \n\x08\x65lements\x18\x01 \x03(\x0b\x32\x0e.sink.v1.Datum"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"8\n\x08Response\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x0f\n\x07\x65rr_msg\x18\x03 \x01(\t"4\n\x0cResponseList\x12$\n\tresponses\x18\x01 \x03(\x0b\x32\x11.sink.v1.Response2\x81\x01\n\x0fUserDefinedSink\x12\x33\n\x06SinkFn\x12\x12.sink.v1.DatumList\x1a\x15.sink.v1.ResponseList\x12\x39\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x16.sink.v1.ReadyResponseB9Z7github.com/numaproj/numaflow-go/pkg/apis/protos/sink/v1b\x06proto3'
)


_EVENTTIME = DESCRIPTOR.message_types_by_name["EventTime"]
_WATERMARK = DESCRIPTOR.message_types_by_name["Watermark"]
_DATUM = DESCRIPTOR.message_types_by_name["Datum"]
_DATUMLIST = DESCRIPTOR.message_types_by_name["DatumList"]
_READYRESPONSE = DESCRIPTOR.message_types_by_name["ReadyResponse"]
_RESPONSE = DESCRIPTOR.message_types_by_name["Response"]
_RESPONSELIST = DESCRIPTOR.message_types_by_name["ResponseList"]
EventTime = _reflection.GeneratedProtocolMessageType(
    "EventTime",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTTIME,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.EventTime)
    },
)
_sym_db.RegisterMessage(EventTime)

Watermark = _reflection.GeneratedProtocolMessageType(
    "Watermark",
    (_message.Message,),
    {
        "DESCRIPTOR": _WATERMARK,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.Watermark)
    },
)
_sym_db.RegisterMessage(Watermark)

Datum = _reflection.GeneratedProtocolMessageType(
    "Datum",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATUM,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.Datum)
    },
)
_sym_db.RegisterMessage(Datum)

DatumList = _reflection.GeneratedProtocolMessageType(
    "DatumList",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATUMLIST,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.DatumList)
    },
)
_sym_db.RegisterMessage(DatumList)

ReadyResponse = _reflection.GeneratedProtocolMessageType(
    "ReadyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _READYRESPONSE,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.ReadyResponse)
    },
)
_sym_db.RegisterMessage(ReadyResponse)

Response = _reflection.GeneratedProtocolMessageType(
    "Response",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESPONSE,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.Response)
    },
)
_sym_db.RegisterMessage(Response)

ResponseList = _reflection.GeneratedProtocolMessageType(
    "ResponseList",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESPONSELIST,
        "__module__": "udsink_pb2"
        # @@protoc_insertion_point(class_scope:sink.v1.ResponseList)
    },
)
_sym_db.RegisterMessage(ResponseList)

_USERDEFINEDSINK = DESCRIPTOR.services_by_name["UserDefinedSink"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z7github.com/numaproj/numaflow-go/pkg/apis/protos/sink/v1"
    _EVENTTIME._serialized_start = 87
    _EVENTTIME._serialized_end = 146
    _WATERMARK._serialized_start = 148
    _WATERMARK._serialized_end = 206
    _DATUM._serialized_start = 208
    _DATUM._serialized_end = 334
    _DATUMLIST._serialized_start = 336
    _DATUMLIST._serialized_end = 381
    _READYRESPONSE._serialized_start = 383
    _READYRESPONSE._serialized_end = 413
    _RESPONSE._serialized_start = 415
    _RESPONSE._serialized_end = 471
    _RESPONSELIST._serialized_start = 473
    _RESPONSELIST._serialized_end = 525
    _USERDEFINEDSINK._serialized_start = 528
    _USERDEFINEDSINK._serialized_end = 657
# @@protoc_insertion_point(module_scope)



import tempfile
import unittest
from datetime import datetime, timezone

from grpc import StatusCode
from grpc_testing import server_from_dictionary, strict_real_time
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as _empty_pb2


from pynumaflow._constants import DATUM_KEY
from pynumaflow.function.generated import udfunction_pb2
from pynumaflow.function.server import UserDefinedFunctionServicer
from pynumaflow.function._dtypes import (
    Message,
    Messages,
    Datum,
)


def map_handler(key: str, datum: Datum) -> Messages:
    val = datum.value
    msg = "payload:%s event_time:%s watermark:%s" % (
        val.decode("utf-8"),
        datum.event_time,
        datum.watermark,
    )
    val = bytes(msg, encoding="utf-8")
    messages = Messages()
    messages.append(Message.to_vtx(key, val))
    return messages


def mock_message():
    msg = bytes("test_mock_message", encoding="utf-8")
    return msg


def mock_event_time():
    t = datetime.fromtimestamp(1662998400, timezone.utc)
    return t


def mock_watermark():
    t = datetime.fromtimestamp(1662998460, timezone.utc)
    return t


class TestServer(unittest.TestCase):
    def __init__(self, method_name) -> None:
        super().__init__(method_name)

        my_servicer = UserDefinedFunctionServicer(map_handler)
        services = {udfunction_pb2.DESCRIPTOR.services_by_name["UserDefinedFunction"]: my_servicer}
        self.test_server = server_from_dictionary(services, strict_real_time())

    def test_is_ready(self):
        method = self.test_server.invoke_unary_unary(
            method_descriptor=(
                udfunction_pb2.DESCRIPTOR.services_by_name["UserDefinedFunction"].methods_by_name[
                    "IsReady"
                ]
            ),
            invocation_metadata={DATUM_KEY, "test"},
            request=_empty_pb2.Empty(),
            timeout=1,
        )

        response, metadata, code, details = method.termination()
        expected = udfunction_pb2.ReadyResponse(ready=True)
        self.assertEqual(expected, response)
        self.assertEqual(code, StatusCode.OK)

    def test_forward_message(self):
        event_time_timestamp = _timestamp_pb2.Timestamp()
        event_time_timestamp.FromDatetime(dt=mock_event_time())
        watermark_timestamp = _timestamp_pb2.Timestamp()
        watermark_timestamp.FromDatetime(dt=mock_watermark())

        request = udfunction_pb2.Datum(
            value=mock_message(),
            event_time=udfunction_pb2.EventTime(event_time=event_time_timestamp),
            watermark=udfunction_pb2.Watermark(watermark=watermark_timestamp),
        )

        method = self.test_server.invoke_unary_unary(
            method_descriptor=(
                udfunction_pb2.DESCRIPTOR.services_by_name["UserDefinedFunction"].methods_by_name[
                    "MapFn"
                ]
            ),
            invocation_metadata={
                (DATUM_KEY, "test"),
                ("this_metadata_will_be_skipped", "test_ignore"),
            },
            request=request,
            timeout=1,
        )

        response, metadata, code, details = method.termination()
        self.assertEqual(1, len(response.elements))
        self.assertEqual("test", response.elements[0].key)
        self.assertEqual(
            bytes(
                "payload:test_mock_message event_time:2022-09-12 16:00:00 watermark:2022-09-12 16:01:00",
                encoding="utf-8",
            ),
            response.elements[0].value,
        )
        self.assertEqual(code, StatusCode.OK)

    def test_reduce_fn(self):
        # TODO: update test after reduce fn is implemented
        event_time_timestamp = _timestamp_pb2.Timestamp()
        event_time_timestamp.FromDatetime(dt=mock_event_time())
        watermark_timestamp = _timestamp_pb2.Timestamp()
        watermark_timestamp.FromDatetime(dt=mock_watermark())

        request = udfunction_pb2.Datum(
            value=mock_message(),
            event_time=udfunction_pb2.EventTime(event_time=event_time_timestamp),
            watermark=udfunction_pb2.Watermark(watermark=watermark_timestamp),
        )

        rpc = self.test_server.invoke_stream_unary(
            method_descriptor=(
                udfunction_pb2.DESCRIPTOR.services_by_name["UserDefinedFunction"].methods_by_name[
                    "ReduceFn"
                ]
            ),
            invocation_metadata={(DATUM_KEY, "test")},
            timeout=1,
        )

        rpc.send_request(request)
        rpc.requests_closed()

        response, metadata, code, details = rpc.termination()
        self.assertIsNone(response)
        self.assertEqual(code, StatusCode.UNKNOWN)
        self.assertEqual("Exception calling application: Method not implemented!", details)


if __name__ == "__main__":
    unittest.main()



import asyncio
import logging
from os import environ


from google.protobuf import empty_pb2 as _empty_pb2

import grpc
from typing import Callable, Any, Iterator, List

from pynumaflow._constants import (
    SINK_SOCK_PATH,
    DATUM_KEY,
)
from pynumaflow.sink import Response, Responses, Datum
from pynumaflow.sink.generated import udsink_pb2_grpc, udsink_pb2
from pynumaflow.types import NumaflowServicerContext

if environ.get("PYTHONDEBUG"):
    logging.basicConfig(level=logging.DEBUG)

_LOGGER = logging.getLogger(__name__)

UDSinkCallable = Callable[[List[Datum], Any], Responses]


class UserDefinedSinkServicer(udsink_pb2_grpc.UserDefinedSinkServicer):
    """
    Provides an interface to write a User Defined Sink (UDSink)
    which will be exposed over gRPC.

    Args:
        sink_handler: Function callable following the type signature of UDSinkCallable
        sock_path: Path to the UNIX Domain Socket

    Example invocation:
    >>> from typing import List
    >>> from pynumaflow.sink import Datum, Responses, Response, UserDefinedSinkServicer
    >>> def udsink_handler(datums: List[Datum]) -> Responses:
    ...   responses = Responses()
    ...   for msg in datums:
    ...     print("User Defined Sink", msg)
    ...     responses.append(Response.as_success(msg.id))
    ...   return responses
    >>> grpc_server = UserDefinedSinkServicer(udsink_handler)
    >>> grpc_server.start()
    """

    def __init__(self, sink_handler: UDSinkCallable, sock_path=SINK_SOCK_PATH):
        self.__sink_handler: UDSinkCallable = sink_handler
        self.sock_path = sock_path
        self._cleanup_coroutines = []

    def SinkFn(
        self, request: udsink_pb2.DatumList, context: NumaflowServicerContext
    ) -> udsink_pb2.ResponseList:
        """
        Applies a sink function to a list of datum elements.
        The pascal case function name comes from the generated udsink_pb2_grpc.py file.
        """

        msgs = self.__sink_handler(request.elements)

        responses = []
        for msg in msgs.items():
            responses.append(udsink_pb2.Response(id=msg.id, success=msg.success, err_msg=msg.err))

        return udsink_pb2.ResponseList(responses=responses)

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> udsink_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the generated udsink_pb2_grpc.py file.
        """
        return udsink_pb2.ReadyResponse(ready=True)

    async def __serve(self) -> None:
        server = grpc.aio.server()
        udsink_pb2_grpc.add_UserDefinedSinkServicer_to_server(
            UserDefinedSinkServicer(self.__sink_handler), server
        )
        server.add_insecure_port(self.sock_path)
        _LOGGER.info("Server listening on: %s", self.sock_path)
        await server.start()

        async def server_graceful_shutdown():
            logging.info("Starting graceful shutdown...")
            # Shuts down the server with 5 seconds of grace period. During the
            # grace period, the server won't accept new connections and allow
            # existing RPCs to continue within the grace period.
            await server.stop(5)

        self._cleanup_coroutines.append(server_graceful_shutdown())
        await server.wait_for_termination()

    def start(self) -> None:
        """Starts the server on the given UNIX socket."""
        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(self.__serve())
        finally:
            loop.run_until_complete(*self._cleanup_coroutines)
            loop.close()



import unittest
from pynumaflow.sink import Response, Responses


class TestResponse(unittest.TestCase):
    def test_as_success(self):
        succ_response = Response.as_success("2")
        self.assertTrue(succ_response.success)

    def test_as_failure(self):
        succ_response = Response.as_failure("3", "RuntimeError encountered!")
        self.assertFalse(succ_response.success)


class TestResponses(unittest.TestCase):
    def setUp(self) -> None:
        self.resps = Responses(
            Response.as_success("2"), Response.as_failure("3", "RuntimeError encountered!")
        )

    def test_responses(self):
        self.resps.append(Response.as_success("4"))
        self.assertEqual(3, len(self.resps.items()))
        self.assertEqual(
            "[Response(id='2', success=True, err=None), "
            "Response(id='3', success=False, err='RuntimeError encountered!'), "
            "Response(id='4', success=True, err=None)]",
            repr(self.resps),
        )

    def test_dumps(self):
        dump_str = self.resps.dumps()
        self.assertEqual(
            "[Response(id='2', success=True, err=None), "
            "Response(id='3', success=False, err='RuntimeError encountered!')]",
            dump_str,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from pynumaflow.function._dtypes import (
    Message,
    Messages,
    ALL,
    DROP,
)


def mock_message():
    msg = bytes("test_mock_message", encoding="utf-8")
    return msg


class TestMessage(unittest.TestCase):
    def test_key(self):
        mock_obj = {"Key": ALL, "Value": mock_message()}
        msgs = Message(key=mock_obj["Key"], value=mock_obj["Value"])
        print(msgs)
        self.assertEqual(mock_obj["Key"], msgs.key)

    def test_value(self):
        mock_obj = {"Key": ALL, "Value": mock_message()}
        msgs = Message(key=mock_obj["Key"], value=mock_obj["Value"])
        self.assertEqual(mock_obj["Value"], msgs.value)

    def test_message_to_all(self):
        mock_obj = {"Key": ALL, "Value": mock_message()}
        msgs = Message.to_all(value=mock_obj["Value"])
        self.assertEqual(Message, type(msgs))
        self.assertEqual(mock_obj["Key"], msgs.key)
        self.assertEqual(mock_obj["Value"], msgs.value)

    def test_message_to_drop(self):
        mock_obj = {"Key": DROP, "Value": b""}
        msgs = Message.to_drop()
        self.assertEqual(Message, type(msgs))
        self.assertEqual(mock_obj["Key"], msgs.key)
        self.assertEqual(mock_obj["Value"], msgs.value)

    def test_message_to(self):
        mock_obj = {"Key": "__KEY__", "Value": mock_message()}
        msgs = Message.to_vtx(key=mock_obj["Key"], value=mock_obj["Value"])
        self.assertEqual(Message, type(msgs))
        self.assertEqual(mock_obj["Key"], msgs.key)
        self.assertEqual(mock_obj["Value"], msgs.value)


class TestMessages(unittest.TestCase):
    @staticmethod
    def mock_message_object():
        value = mock_message()
        return Message(ALL, value)

    def test_items(self):
        mock_obj = [
            {"Key": b"U+005C__ALL__", "Value": mock_message()},
            {"Key": b"U+005C__ALL__", "Value": mock_message()},
        ]
        msgs = Messages()
        msgs._messages = mock_obj
        self.assertEqual(len(mock_obj), len(msgs.items()))
        self.assertEqual(mock_obj[0]["Key"], msgs.items()[0]["Key"])
        self.assertEqual(mock_obj[0]["Value"], msgs.items()[0]["Value"])
        self.assertEqual(
            "[{'Key': b'U+005C__ALL__', 'Value': b'test_mock_message'}, "
            "{'Key': b'U+005C__ALL__', 'Value': b'test_mock_message'}]",
            repr(msgs),
        )

    def test_append(self):
        msgs = Messages()
        self.assertEqual(0, len(msgs.items()))
        msgs.append(self.mock_message_object())
        self.assertEqual(1, len(msgs.items()))
        msgs.append(self.mock_message_object())
        self.assertEqual(2, len(msgs.items()))

    def test_message_forward(self):
        mock_obj = Messages()
        msg = self.mock_message_object()
        mock_obj.append(msg)
        true_obj = Messages.as_forward_all(mock_message())
        self.assertEqual(type(mock_obj), type(true_obj))
        for i in range(len(true_obj.items())):
            self.assertEqual(type(mock_obj.items()[i]), type(true_obj.items()[i]))
            self.assertEqual(mock_obj.items()[i].key, true_obj.items()[i].key)
            self.assertEqual(mock_obj.items()[i].value, true_obj.items()[i].value)

    def test_message_forward_to_drop(self):
        mock_obj = Messages()
        mock_obj.append(Message(DROP, bytes()))
        true_obj = Messages.as_forward_all(bytes())
        self.assertEqual(type(mock_obj), type(true_obj))
        for i in range(len(true_obj.items())):
            self.assertEqual(type(mock_obj.items()[i]), type(true_obj.items()[i]))
            self.assertEqual(mock_obj.items()[i].key, true_obj.items()[i].key)
            self.assertEqual(mock_obj.items()[i].value, true_obj.items()[i].value)

    def test_dump(self):
        msgs = Messages()
        msgs.append(self.mock_message_object())
        msgs.append(self.mock_message_object())
        self.assertEqual(
            "[{b'U+005C__ALL__': b'test_mock_message'}, {b'U+005C__ALL__': b'test_mock_message'}]",
            msgs.dumps(),
        )

    def test_load(self):
        # to improve codecov
        msgs = Messages()
        msgs.loads()


if __name__ == "__main__":
    unittest.main()



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from typing import (
    ClassVar as _ClassVar,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
    List,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ReadyResponse(_message.Message):
    __slots__ = ["ready"]
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: _Optional[bool] = ...) -> None: ...

class EventTime(_message.Message):
    __slots__ = ["event_time"]
    EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
    event_time: _timestamp_pb2.Timestamp
    def __init__(self, event_time: _Optional[_timestamp_pb2.Timestamp] = ...) -> None: ...

class Watermark(_message.Message):
    __slots__ = ["watermark"]
    WATERMARK_FIELD_NUMBER: _ClassVar[int]
    watermark: _timestamp_pb2.Timestamp
    def __init__(self, watermark: _Optional[_timestamp_pb2.Timestamp] = ...) -> None: ...

class Datum(_message.Message):
    __slots__ = ["key", "value", "event_time", "watermark"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
    WATERMARK_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: bytes
    event_time: _timestamp_pb2.Timestamp
    watermark: _timestamp_pb2.Timestamp
    def __init__(
        self,
        key: _Optional[str],
        value: _Optional[bytes],
        event_time: _Optional[_timestamp_pb2.Timestamp] = ...,
        watermark: _Optional[_timestamp_pb2.Timestamp] = ...,
    ) -> None: ...

class DatumList(_message.Message):
    __slots__ = ["elements"]
    ELEMENTS_FIELD_NUMBER: _ClassVar[int]
    elements: List[Datum]
    def __init__(self, elements: _Optional[List[Datum]]) -> None: ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pynumaflow',
 'pynumaflow.function',
 'pynumaflow.function.generated',
 'pynumaflow.sink',
 'pynumaflow.sink.generated',
 'pynumaflow.tests',
 'pynumaflow.tests.function',
 'pynumaflow.tests.sink']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'grpcio-tools>=1.48.1,<2.0.0',
 'grpcio>=1.48.1,<2.0.0',
 'msgpack>=1.0.3,<2.0.0']

setup_kwargs = {
    'name': 'pynumaflow',
    'version': '0.2.1',
    'description': 'Provides the interfaces of writing Python User Defined Functions and Sinks for NumaFlow.',
    'long_description': '# Python SDK for Numaflow\n\nThis SDK provides the interface for writing [UDFs](https://numaproj.github.io/numaflow/user-defined-functions/) \nand [UDSinks](https://numaproj.github.io/numaflow/sinks/user-defined-sinks/) in Python.\n\n## Implement a User Defined Function (UDF)\n\n```python\n\nfrom pynumaflow.function import Messages, Message, Datum, UserDefinedFunctionServicer\n\n\ndef map_handler(key: str, datum: Datum) -> Messages:\n    val = datum.value\n    _ = datum.event_time\n    _ = datum.watermark\n    messages = Messages()\n    messages.append(Message.to_vtx(key, val))\n    return messages\n\n\nif __name__ == "__main__":\n    grpc_server = UserDefinedFunctionServicer(map_handler)\n    grpc_server.start()\n```\n\n### Sample Image (TODO)\n\n## Implement a User Defined Sink (UDSink)\n\n```python\nfrom typing import List\nfrom pynumaflow.sink import Message, Responses, Response, HTTPSinkHandler\n\n\ndef udsink_handler(messages: List[Message], __) -> Responses:\n    responses = Responses()\n    for msg in messages:\n        responses.append(Response.as_success(msg.id))\n    return responses\n\n\nif __name__ == "__main__":\n    handler = HTTPSinkHandler(udsink_handler)\n    handler.start()\n```\n\n### Sample Image\n\nA sample UDSink [Dockerfile](examples/sink/simplesink/Dockerfile) is provided \nunder [examples](examples/sink/simplesink).\n',
    'author': 'NumaFlow Developers',
    'author_email': 'None',
    'maintainer': 'Avik Basu',
    'maintainer_email': 'avikbasu93@gmail.com',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from pynumaflow.function.generated import udfunction_pb2 as udfunction__pb2


class UserDefinedFunctionStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.MapFn = channel.unary_unary(
            "/function.v1.UserDefinedFunction/MapFn",
            request_serializer=udfunction__pb2.Datum.SerializeToString,
            response_deserializer=udfunction__pb2.DatumList.FromString,
        )
        self.ReduceFn = channel.stream_unary(
            "/function.v1.UserDefinedFunction/ReduceFn",
            request_serializer=udfunction__pb2.Datum.SerializeToString,
            response_deserializer=udfunction__pb2.DatumList.FromString,
        )
        self.IsReady = channel.unary_unary(
            "/function.v1.UserDefinedFunction/IsReady",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=udfunction__pb2.ReadyResponse.FromString,
        )


class UserDefinedFunctionServicer(object):
    """Missing associated documentation comment in .proto file."""

    def MapFn(self, request, context):
        """Applies a function to each datum element."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ReduceFn(self, request_iterator, context):
        """Applies a reduce function to a datum stream."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def IsReady(self, request, context):
        """IsReady is the heartbeat endpoint for gRPC."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_UserDefinedFunctionServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "MapFn": grpc.unary_unary_rpc_method_handler(
            servicer.MapFn,
            request_deserializer=udfunction__pb2.Datum.FromString,
            response_serializer=udfunction__pb2.DatumList.SerializeToString,
        ),
        "ReduceFn": grpc.stream_unary_rpc_method_handler(
            servicer.ReduceFn,
            request_deserializer=udfunction__pb2.Datum.FromString,
            response_serializer=udfunction__pb2.DatumList.SerializeToString,
        ),
        "IsReady": grpc.unary_unary_rpc_method_handler(
            servicer.IsReady,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=udfunction__pb2.ReadyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "function.v1.UserDefinedFunction", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class UserDefinedFunction(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def MapFn(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/function.v1.UserDefinedFunction/MapFn",
            udfunction__pb2.Datum.SerializeToString,
            udfunction__pb2.DatumList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ReduceFn(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            "/function.v1.UserDefinedFunction/ReduceFn",
            udfunction__pb2.Datum.SerializeToString,
            udfunction__pb2.DatumList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def IsReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/function.v1.UserDefinedFunction/IsReady",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            udfunction__pb2.ReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



APPLICATION_JSON = "application/json"
APPLICATION_MSG_PACK = "application/msgpack"
NUMAFLOW_UDF_CONTENT_TYPE = "NUMAFLOW_UDF_CONTENT_TYPE"
NUMAFLOW_MESSAGE_KEY = "x-numa-message-key"
FUNCTION_SOCK_PATH = "/var/run/numaflow/udf.sock"
SINK_SOCK_PATH = "/var/run/numaflow/udsink.sock"
# TODO: need to make sure the DATUM_KEY value is the same as
# https://github.com/numaproj/numaflow-go/blob/main/pkg/function/configs.go#L6
DATUM_KEY = "x-numaflow-datum-key"
